/* +------------------------------------------------------------------------+
   |                                                                        |
   |                         Racine carre rcursive                        |
   |                                                                        |
   +------------------------------------------------------------------------+ */

#include "defs.h"

static ndouble un = 1;

/* c <- 2*trunc(a^1/2), lg(a) = 2n, lg(c) = n                       */
/* si rem = 0 alors a <- a - high(c^2/4) sinon a <- a - c^2/4       */
/* si rem = 0 le quotient est trop grand d'au plus 2+2log(n) units */
/* le chiffre de poids fort de a doit tre compris entre            */
/* base/16 inclus et base/4 exclu                                   */

void zimmermann(int rem, ndouble *c, ndouble *a, int n, int n0) {

  if (n <= n0) sqrtn2(c,a,n);
  else {
    int p=n/2, q=n-p;
    ndouble *buff = getmem(n);
    zdouble z;

    /* racine carre de la partie haute */
    zimmermann(1,c+p, a+2*p,q,n0);

    /* partie basse par division tronque */
    if (compare(a+n,c+q,p)< 0) burnikel(rem,c,a+q,c+q,p,n0);
    else {
      int i,j;

      for (i=0; i<p; i++) c[i] = mask;
      clearmem(a+n,p);
      for (i=p; i>0;) {
	j = ((i <= n0) || (rem)) ? i : (i+1)/2;
	inc(a+n-j,c+n-i,j+1,j);
	i -= j;
      }
    }

    if (rem) {

      /* rectangle q x p -> carr + bande ventuelle */
      karasqr(buff,c,p,n0);

      if (q > p) {
	int i;
	ndouble r;

	for (i=0, r=0; i < p; i++) {
	  r += c[p]*c[i] + buff[i+p];
	  buff[i+p] = r & mask;
	  r >>= shift;
	}
	buff[n-1] = r;
      }

      /* double la partie basse et retire le rectangle */
      inc(c,c,p+1,p);
      z = dec(a,buff,n+p+1,n);

      while (z) {dec(c,&un,p+1,1); z += inc(a,c,n+p+1,n); dec(c,&un,p+1,1);}
    }

    freemem(buff);
  }
}
    
/* corrige la racine sans reste */
void zimmermann_correct(ndouble *c, ndouble *a, int n, int n0) {

  if (n > n0) {
    int p=n/2, q=n-p;
    ndouble *buff = getmem(n);
    zdouble z;

    /* corrige le reste de la deuxime division */
    burnikel_correct(c,a+q,c+q,p,n0);

    /* double la partie basse et retire le rectangle */
    karasqr(buff,c,p,n0);

    if (q > p) {
      int i;
      ndouble r;
      
      for (i=0, r=0; i < p; i++) {
	r += c[p]*c[i] + buff[i+p];
	buff[i+p] = r & mask;
	r >>= shift;
      }
      buff[n-1] = r;
    }
    
    inc(c,c,p+1,p);
    z = dec(a,buff,n+p+1,n);
    while (z) {dec(c,&un,p+1,1); z += inc(a,c,n+p+1,n); dec(c,&un,p+1,1);}

    freemem(buff);
  }
}

                    /* +----------------------------+
                       |  Racine carre sans reste  |
                       +----------------------------+ */

void zimmermann_nr(ndouble *c, ndouble *a, int n, int n0) {

  if (n <= n0) sqrtn2(c,a,n);
  else {

    /* calcule la racine sans reste avec deux chiffres aprs la virgule */
    ndouble *a1 = getmem(2*n+4);
    ndouble *c1 = getmem(n+2);

    movemem(a1+4,a,2*n); clearmem(a1,4);
    zimmermann(0,c1,a1,n+2,n0);

    /* si la partie dcimale est trop petite corrige la division */
    if (c1[1] == 0) zimmermann_correct(c1,a1,n+2,n0);
    else inc(c1,c1,n/2+2,n/2+1);

    /* tronque le rsultat */
    movemem(c,c1+2,n); c[0] = c[0] & (mask-1);

    freemem(a1); freemem(c1);
  }
}

/* tests */

void sqrtsz(int n, int *sa, int *sb, int *sc, int *sd) {
  *sa = 2*n;
  *sb = -n;
  *sc = -2*n;
  *sd = 0;
}

void zimm0(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {

  a[2*n-1] = (a[2*n-1] & (mask/4)) | (msb/4);
  movemem(c, a, 2*n);
  zimmermann_nr(b, c, n, n0);

}

void zimm1(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {

  a[2*n-1] = (a[2*n-1] & (mask/4)) | (msb/4);
  movemem(c, a, 2*n);
  zimmermann(1, b, c, n, n0);

}

int sqrtnrchk(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {
  ndouble *e, *f, r;
  int res;

  e = getmem(n); f = getmem(2*n);
  div2(e, b, n);

  karasqr(f, e, n, n0);
  res = (compare(f, a, 2*n) > 0);
  r = inc(f, b, 2*n, n) + inc(f, &un, 2*n, 1);
  res |= (!r) && (compare(f, a, 2*n) <= 0);

  freemem(e); freemem(f);
  return(res);
}

int sqrtchk(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {
  ndouble *e, *f, r;
  int res;

  e = getmem(n); f = getmem(2*n);
  div2(e, b, n);

  karasqr(f, e, n, n0);
  r = inc(f, c, 2*n, n);
  res = (r) || (compare(f, a, 2*n)) || (compare(c, b, n) >= 0);

  freemem(e); freemem(f);
  return(res);
}
